import zmq from 'zeromq/v5-compat.js';
import * as ConfigContainer from '../../lib/config.cjs';
import { ZMQ_PUB_CHANNELS } from '../../lib/system/constants.mjs';

export class ZmqObserver {
    /**
     * @typedef {Class} ZmqObserver
     */

    /**
     * @property {Object|null}
     */
    #socket = null;

    /**
     * @property {Array<Object>}
     */
    #subscriptions = [];

    constructor() {
        this.#socket = zmq.socket('sub'); // subscription to rtls-pub-sub-proxyd 7001 port
        this.#socket.connect(ConfigContainer.config.zmqPubSubProxydUrl);
        this.#subscriptions = [];
    }

    /**
     * @method
     * @return {VoidFunction}
     */
    init() {
        this.#socket.subscribe('');
        this.#socket.on('message', this.handleMessage);
    }

    /**
     * @method
     * @return {VoidFunction}
     */
    destroy() {
        this.#socket.close();
    }

    /**
     * @static
     * @param {Buffer} bufferMsg - Buffer from ZMQ
     * @return {Array<any>|null}
     */
    static parseZmqMsg(bufferMsg) {
        const message = bufferMsg.toString();
        if (message.startsWith(`${ZMQ_PUB_CHANNELS.EVENT} `)) {
            const result = JSON.parse(message.replace(`${ZMQ_PUB_CHANNELS.EVENT} `, ''));
            return [result, ZMQ_PUB_CHANNELS.EVENT];
        } else if (message.startsWith(`${ZMQ_PUB_CHANNELS.POS} `)) {
            const result = JSON.parse(message.replace(`${ZMQ_PUB_CHANNELS.POS} `, ''));
            return [result, ZMQ_PUB_CHANNELS.POS];
        } else if (message.startsWith(`${ZMQ_PUB_CHANNELS.DATA} `)) {
            const result = JSON.parse(message.replace(`${ZMQ_PUB_CHANNELS.DATA} `, ''));
            return [result, ZMQ_PUB_CHANNELS.DATA];
        } else if (message.startsWith(`${ZMQ_PUB_CHANNELS.OBSOLETE} `)) {
            const result = JSON.parse(message.replace(`${ZMQ_PUB_CHANNELS.OBSOLETE} `, ''));
            return [result, ZMQ_PUB_CHANNELS.OBSOLETE];
        }
        return null;
    }

    /**
     * @method
     * @param {Buffer} bufferMsg - Buffer from ZMQ
     * @return {Boolean}
     */
    handleMessage = (bufferMsg) => {
        const [zmqMsg, channel] = ZmqObserver.parseZmqMsg(bufferMsg);
        const subscription = this.#findSubscription(channel);
        if (subscription) {
            this.#unsubscribe(channel);
            subscription.callback(zmqMsg, channel);
        }
    };

    /**
     * @method
     * @param {Object} channel - ZMQ channel
     * @return {Object}
     */
    #findSubscription(channel) {
        return this.#subscriptions.find((subscription) => subscription.channel === channel);
    }

    /**
     * @method
     * @param {Object} channel - ZMQ channel
     * @return {VoidFunction}
     */
    #unsubscribe(channel) {
        this.#subscriptions = this.#subscriptions.filter((subscription) => subscription.channel !== channel);
    }

    /**
     * @method
     * @param {String} channel - subscription channel
     * @param {Function} callback - callback
     * @return {VoidFunction}
     */
    async subscribe(channel = '', callback = () => {}) {
        this.#subscriptions.push({ channel, callback });
    }
}
