import UseCaseBase from '../UseCaseBase.mjs';
import { Device, NodeAssociation, Event, Sublocation, Position } from '../../models/index.mjs';
import { formatHeartbeat } from './formatters.mjs';
import { formatPosition } from '../position/formatters.mjs';
import { UID } from '../../utils/index.mjs';
import logger from '../../api/logger.mjs';

export class EventsHandle extends UseCaseBase {
    /**
     * @method execute
     * @param {Object} params  - raw event form JSON/BIN convertor
     * @property {String} params.ts  - position timestamp in ISO 8601 format
     * @property {String} params.mac  - hardware device MAC address
     * @property {String} params.type  - message type [announce|heartbeat]
     * @property {Object} params.data  - position properties
     * @property {Number} params.data.x  - the X coordinate of the position
     * @property {Number} params.data.y  - the Y coordinate of the position
     * @property {Number} params.data.z  - the Z coordinate of the position
     * @property {Number} params.data.sublocation_id  - sublocation numeric identifier
     * @returns {Promise<null>}
     */
    async execute(params) {
        const association = await new NodeAssociation().findByMac(params);
        if (!association) {
            // auto create ?
            //
            const newUid = UID();
            const device = await new Device().findByMacOrCreate(params.mac, params.ts, newUid);
            if (device.uid === newUid) {
                return null;
            }
        }
        const obj_uid = association ? association.anchor_uid || association.trackable_uid : null;
        const isWin = association ? !!association.anchor_uid : false;
        if (obj_uid) {
            const sublocationUid = await new Sublocation().findUidById({
                sl_id: params.data.sublocation_id
            });
            switch (params.type) {
                case 'heartbeat':
                case 'announce': {
                    const heartbeat = formatHeartbeat(params, obj_uid, sublocationUid, isWin);
                    await this.notificator.notify(Event.notifyMessage(heartbeat));
                    if (!isWin && sublocationUid) {
                        const pos = formatPosition(params, obj_uid, sublocationUid, 'HB');
                        await this.notificator.notify(Position.notifyMessage(pos));
                    }
                }
                    break;
                default: {
                    logger.error(`Unknown event type: ${params.type}`);
                }
                    break;
            }
        }
        await new Device().updateLastHeard(params.mac, params.ts);

        return null;
    }
}
