import assert from 'assert';
import _cloneDeep from 'lodash.clonedeep';
import { UID } from '../../lib/utils/index.mjs';
import { ERROR_CODE, ZMQ_ACTIONS, ZMQ_CHANNELS, ZMQ_OBJECTS } from '../../lib/system/constants.mjs';
import { isISOTimeFormat, ZmqObserver, get, post, del } from '../utils/index.mjs';
import { createLocation, checkEntities, createLsbSchedule, Server } from './helpers.mjs';

const location1 = {};
const location2 = {};
const location3 = {};
const location4 = {};
const location5 = {};
const setting1 = {};
const setting2 = {};
const schedule = {};

const settingBody = {
    value: {
        active_schedule_uid: 'tGhW9YSET-GoTy2FI5TzsA',
        lsb: 'ON',
        masters: [{
            anchor_uid: 'MASTER_1-GhW9YSET-GoTy',
            sfd: 500,
            auto_sync: true
        }],
        distributors: [{
            anchor_uid: 'DISTRIB_1-GhW9YSET-GoT',
            master_uid: null,
            slot: 0
        }]
    }
};

describe('Network Settings Management API', function() {
    before(async function() {
        await Server().init();
        Object.assign(location1, await createLocation('Location Name1'));
        Object.assign(location2, await createLocation('Location Name2'));
        Object.assign(location3, await createLocation('Location Name3'));
        Object.assign(location4, await createLocation('Location Name4'));
        Object.assign(location5, await createLocation('Location Name5'));
        Object.assign(schedule, await createLsbSchedule(location1.uid, 'LSB schedule#1'));
    });

    after(async function() {
        await del(`/api/v3/locations/${location1.uid}`);
        await del(`/api/v3/locations/${location2.uid}`);
        await del(`/api/v3/locations/${location3.uid}`);
        await del(`/api/v3/locations/${location4.uid}`);
        await del(`/api/v3/locations/${location5.uid}`);
        await del(`/api/v3/network/lsb_schedules/${schedule.uid}`);
    });

    describe('Get Empty Network Setting', function() {
        it('should return empty list of settings', checkEntities('network/settings', [], 'settings'));

        it('should return NotFound error on wrong location uid', function(done) {
            const invalidUid = UID();
            get(`/api/v3/network/settings/${invalidUid}`)
                .expect(404)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        `Network settings for location with uid ${invalidUid} not found`,
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Save Network Setting', function() {
        describe('Negative cases', function() {
            it('should fail save setting for non existed location', function(done) {
                const body = _cloneDeep(settingBody);
                const invalidUid = UID();
                body.location_uid = invalidUid;
                body.value.active_schedule_uid = schedule.uid;
                post('/api/v3/network/settings', body)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Location with uid ${invalidUid} not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting for non existed schedule', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                const invalidUid = UID();
                body.value.active_schedule_uid = invalidUid;
                post('/api/v3/network/settings', body)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `LSB schedule with uid ${invalidUid} not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting without location_uid', function(done) {
                const body = _cloneDeep(settingBody);
                delete body.location_uid;
                body.value.active_schedule_uid = schedule.uid;
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"location_uid":"REQUIRED"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with empty value object', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value = {};
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"active_schedule_uid":"REQUIRED","lsb":"REQUIRED","masters":"REQUIRED","distributors":"REQUIRED"}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with empty active_schedule_uid', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                delete body.value.active_schedule_uid;
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"active_schedule_uid":"REQUIRED"}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });

            it('should fail save setting with wrong master format. Empty object', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.masters = [{}];
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"masters":[{"anchor_uid":"REQUIRED","sfd":"REQUIRED","auto_sync":"REQUIRED"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with wrong master format. Wrong sfd.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.masters = [{ anchor_uid: UID(), sfd: -100, auto_sync: true }];
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"masters":[{"sfd":"NOT_ALLOWED_VALUE"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting without masters', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location3.uid;
                body.value.active_schedule_uid = schedule.uid;
                delete body.value.masters;
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"masters":"REQUIRED"}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with empty masters.anchor_uid format.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.masters[0].anchor_uid = '';
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"masters":[{"anchor_uid":"CANNOT_BE_EMPTY"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting without masters.anchor_uid format.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                delete body.value.masters[0].anchor_uid;
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"masters":[{"anchor_uid":"REQUIRED"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });

            it('should fail save setting with masters wrong body.distributors', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location3.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.distributors = [{
                    master_uid: 'DISTRE_1-GhW9YSET-GoTy',
                    slot: 150
                }, {
                    anchor_uid: 'DISTRIB_2-GhW9YSET-GoT',
                    master_uid: 'MASTER_1-GhW9YSET-GoTy',
                    slot: 151
                }];
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"distributors":[{"anchor_uid":"REQUIRED"},null]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with wrong distributor format. Empty object.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.distributors = [{}];
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"distributors":[{"anchor_uid":"REQUIRED","master_uid":"REQUIRED","slot":"REQUIRED"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting without distributors', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location4.uid;
                body.value.active_schedule_uid = schedule.uid;
                delete body.value.distributors;
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"distributors":"REQUIRED"}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with empty distributors.anchor_uid format.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.distributors[0].anchor_uid = '';
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"distributors":[{"anchor_uid":"CANNOT_BE_EMPTY"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting without distributors.anchor_uid format.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                delete body.value.distributors[0].anchor_uid;
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"distributors":[{"anchor_uid":"REQUIRED"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with empty distributors.master_uid format.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.distributors[0].master_uid = '';
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"distributors":[{"master_uid":"CANNOT_BE_EMPTY"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting without distributors.master_uid format.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                delete body.value.distributors[0].master_uid;
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"distributors":[{"master_uid":"REQUIRED"}]}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
    
            it('should fail save setting with no lsb.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                delete body.value.lsb;
                body.value.active_schedule_uid = schedule.uid;
                body.value.masters = [{ anchor_uid: UID(), sfd: 500, auto_sync: true }];
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"lsb":"REQUIRED"}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with empty lsb.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.lsb = '';
                body.value.active_schedule_uid = schedule.uid;
                body.value.masters = [{ anchor_uid: UID(), sfd: 1000, auto_sync: true }];
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"lsb":"CANNOT_BE_EMPTY"}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail save setting with wrong lsb.', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.lsb = 'wrong';
                body.value.active_schedule_uid = schedule.uid;
                body.value.masters = [{ anchor_uid: UID(), sfd: 250, auto_sync: true }];
                post('/api/v3/network/settings', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"value":{"lsb":"NOT_ALLOWED_VALUE"}}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
        });

        describe('Positive cases', function() {
            it('should save setting with empty masters list', function (done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.masters = [];
                post('/api/v3/network/settings', body)
                    .expect(200)
                    .then((response) => {
                        assert(response.body.location_uid === location1.uid, `Wrong location "uid"`);
                        assert(isISOTimeFormat(response.body.created_at), `Wrong setting "created_at" format`);
                        assert(isISOTimeFormat(response.body.updated_at), `Wrong setting "updated_at" format`);
    
                        assert.strictEqual(
                            response.body.value.active_schedule_uid,
                            body.value.active_schedule_uid,
                            `Wrong setting active schedule uid value`
                        );
                        assert.deepStrictEqual(response.body.value, { ...body.value, tdoa_version: 3 }, `Wrong "value" value`);
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should save setting with empty distributors list', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                body.value.distributors = [];
                post('/api/v3/network/settings', body)
                    .expect(200)
                    .then((response) => {
                        assert(response.body.location_uid === location1.uid, `Wrong location "uid"`);
                        assert(isISOTimeFormat(response.body.created_at), `Wrong setting "created_at" format`);
                        assert(isISOTimeFormat(response.body.updated_at), `Wrong setting "updated_at" format`);
    
                        assert.strictEqual(
                            response.body.value.active_schedule_uid,
                            body.value.active_schedule_uid,
                            `Wrong setting active schedule uid value`
                        );
                        assert.deepStrictEqual(response.body.value, { ...body.value, tdoa_version: 3 }, `Wrong "value" value`);
                        done();
                    })
                    .catch((err) => done(err));
            });
            
            it('should create Network setting #1', function(done) {
                const body = _cloneDeep(settingBody);
                body.value.lsb = 'on';
                body.value.active_schedule_uid = schedule.uid;
                body.location_uid = location1.uid;
                post('/api/v3/network/settings', body)
                    .expect(200)
                    .then((response) => {
                        assert(response.body.location_uid === location1.uid, `Wrong location "uid"`);
                        assert(isISOTimeFormat(response.body.created_at), `Wrong setting "created_at" format`);
                        assert(isISOTimeFormat(response.body.updated_at), `Wrong setting "updated_at" format`);
                        assert(body.value.lsb.toUpperCase() === response.body.value.lsb, `Wrong setting "lsb" format`);

                        assert.strictEqual(
                            response.body.value.active_schedule_uid,
                            body.value.active_schedule_uid,
                            `Wrong setting active schedule uid value`
                        );
                        body.value.lsb = body.value.lsb.toUpperCase();
                        assert.deepStrictEqual(response.body.value, { ...body.value, tdoa_version: 3 }, `Wrong "value" value`);
                        Object.assign(setting1, response.body);
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should update Network setting #1', function(done) {
                const body = _cloneDeep(settingBody);
                body.value.lsb = 'off';
                body.location_uid = location1.uid;
                body.value.active_schedule_uid = schedule.uid;
                post('/api/v3/network/settings', body)
                    .expect(200)
                    .then((response) => {
                        assert(response.body.location_uid === location1.uid, `Wrong location "uid"`);
                        assert(isISOTimeFormat(response.body.created_at), `Wrong setting "created_at" format`);
                        assert(isISOTimeFormat(response.body.updated_at), `Wrong setting "updated_at" format`);
                        assert(body.value.lsb.toUpperCase() === response.body.value.lsb, `Wrong setting "lsb" format`);
                        
                        assert(
                            new Date(response.body.updated_at) > new Date(response.body.created_at),
                            `Wrong setting "updated_at" value`
                        );

                        assert.strictEqual(
                            response.body.value.active_schedule_uid,
                            body.value.active_schedule_uid,
                            `Wrong setting active schedule uid value`
                        );
                        body.value.lsb = body.value.lsb.toUpperCase();
                        assert.deepStrictEqual(response.body.value, { ...body.value, tdoa_version: 3 }, `Wrong "value" value`);
                        Object.assign(setting1, response.body);
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should create Network setting #2', function(done) {
                const body = _cloneDeep(settingBody);
                body.location_uid = location2.uid;
                body.value.active_schedule_uid = schedule.uid;
                post('/api/v3/network/settings', body)
                    .expect(200)
                    .then((response) => {
                        assert(response.body.location_uid === location2.uid, `Wrong location "uid"`);
                        assert(isISOTimeFormat(response.body.created_at), `Wrong setting "created_at" format`);
                        assert(isISOTimeFormat(response.body.updated_at), `Wrong setting "updated_at" format`);
                        assert(
                            new Date(response.body.created_at).getTime() ===
                                new Date(response.body.updated_at).getTime(),
                            `The "updated_at" and "created_at" values not equal`
                        );

                        assert.strictEqual(
                            response.body.value.active_schedule_uid,
                            body.value.active_schedule_uid,
                            `Wrong setting active schedule uid value`
                        );
                        assert.deepStrictEqual(response.body.value, { ...body.value, tdoa_version: 3 }, `Wrong "value" value`);
                        Object.assign(setting2, response.body);
                        done();
                    })
                    .catch((err) => done(err));
            });
        });
    });

    describe('Get Network Settings', function() {
        it('should return list of settings', function(done) {
            const settingsList = [setting1, setting2];
            get(`/api/v3/network/settings`)
                .expect(200)
                .then((response) => {
                    assert(response.body.settings.length, 2, `Network settings number not equal to 2`);

                    for (const settingRes of response.body.settings) {
                        const settingReq = settingsList.find(setting => setting.location_uid === settingRes.location_uid);
                        assert.deepStrictEqual(settingReq, { ...settingRes, value: { tdoa_version: 3, ...settingRes.value }}, 'Settings are not equal');
                    }
                    done();
                })
                .catch((err) => done(err));
        });
        it('should return setting by location uid #1', function(done) {
            get(`/api/v3/network/settings/${setting1.location_uid}`)
                .expect(200)
                .then((response) => {
                    assert(response.body.settings.length === 1, `Network settings number not equal to 1`);
                    assert.deepStrictEqual(response.body.settings[0], setting1, 'Settings are not equal');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should return setting by location uid #2', function(done) {
            get(`/api/v3/network/settings/${setting2.location_uid}`)
                .expect(200)
                .then((response) => {
                    assert(response.body.settings.length === 1, `Network settings number not equal to 1`);
                    assert.deepStrictEqual(response.body.settings[0], setting2, 'Settings are not equal');
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Delete LSB schedule', function() {
        describe('Negative Cases', function() {
            it('should return Not Found error on delete setting for non existing location', function(done) {
                const invalidUid = UID();
                del(`/api/v3/network/settings/${invalidUid}`)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Location with uid ${invalidUid} not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should return Not Found error on delete non existing setting', function(done) {
                del(`/api/v3/network/settings/${location5.uid}`)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Network settings for location with uid ${location5.uid} not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
        });

        describe('Remove Setting on Location removal (#2)', function() {
            before(async function() {
                await del(`/api/v3/locations/${location2.uid}`);
            });

            it('should return Not Found error on reading setting for remove location settings', function(done) {
                get(`/api/v3/network/settings/${location2.uid}`)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Network settings for location with uid ${location2.uid} not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
        });

        describe('Positive Cases  (#1, #3, #4)', function() {
            it('should successfully delete setting #1', function(done) {
                del(`/api/v3/network/settings/${setting1.location_uid}`)
                    .expect(200)
                    .then((response) => {
                        assert.deepStrictEqual(response.body, { location_uid: setting1.location_uid }, `Wrong setting value`);
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should return empty settings', function(done) {
                get(`/api/v3/network/settings`)
                    .expect(200)
                    .then((response) => {
                        assert.deepStrictEqual(
                            response.body,
                            { settings: [] },
                            `Network settings number not equal to 0`
                        );
                        done();
                    })
                    .catch((err) => done(err));
            });
        });
    });

    describe('ZMQ Messaging for Network Settings CRUD', function() {
        const zmqObserver = new ZmqObserver();
        const networkSettingZmq = {};

        before(function() {
            zmqObserver.init();
        });
        after(function() {
            zmqObserver.destroy();
        });

        it('should send zmq created network setting message', function(done) {
            const zmqSubscription = { action: ZMQ_ACTIONS.CREATED, object: ZMQ_OBJECTS.NETWORK_SETTING };
            const body = _cloneDeep(settingBody);
            body.location_uid = location1.uid;
            body.value.active_schedule_uid = schedule.uid;

            const subscription = zmqObserver.subscribe(zmqSubscription, function(zmqMsg, channel) {
                assert(isISOTimeFormat(zmqMsg.ts), `Wrong ZMQ message "ts" format`);
                assert(channel === ZMQ_CHANNELS.EVENT, 'Wrong ZMQ message channel');
                assert.deepStrictEqual(zmqMsg.data.value, { ...body.value, tdoa_version: 3 }, 'ZMQ message has wrong data');
                assert.strictEqual(zmqMsg.data.location_uid, body.location_uid, 'ZMQ message has wrong location_uid');
                assert(isISOTimeFormat(zmqMsg.data.created_at), `Wrong ZMQ message "created_at" format`);
                assert(isISOTimeFormat(zmqMsg.data.updated_at), `Wrong ZMQ message "updated_at" format`);
                done();
            });
            subscription
                .then(() => post('/api/v3/network/settings', body))
                .then((response) => {
                    Object.assign(networkSettingZmq, response.body);
                });
        });

        it('should send zmq deleted network setting message', function(done) {
            const zmqSubscription = { action: ZMQ_ACTIONS.DELETED, object: ZMQ_OBJECTS.NETWORK_SETTING };

            const subscription = zmqObserver.subscribe(zmqSubscription, function(zmqMsg, channel) {
                assert(isISOTimeFormat(zmqMsg.ts), `Wrong ZMQ message "ts" format`);
                assert(channel === ZMQ_CHANNELS.EVENT, 'Wrong ZMQ message channel');
                assert.deepStrictEqual(zmqMsg.data.value, networkSettingZmq.value, 'ZMQ message has wrong data');
                assert.strictEqual(
                    zmqMsg.data.location_uid,
                    networkSettingZmq.location_uid,
                    'ZMQ message has wrong location_uid'
                );
                assert(isISOTimeFormat(zmqMsg.data.created_at), `Wrong ZMQ message "created_at" format`);
                assert(isISOTimeFormat(zmqMsg.data.updated_at), `Wrong ZMQ message "updated_at" format`);
                done();
            });
            subscription.then(() => del(`/api/v3/network/settings/${location1.uid}`));
        });
    });
});
