import assert from 'assert';
import { UID } from '../../lib/utils/index.mjs';
import { ERROR_CODE, ZMQ_ACTIONS, ZMQ_CHANNELS, ZMQ_OBJECTS } from '../../lib/system/constants.mjs';
import { isISOTimeFormat, isUpperCaseString, ZmqObserver, get, post, put, del } from '../utils/index.mjs';
import {
    createTrackable,
    createAnchor,
    createNodeAssociation,
    createNode,
    createTag,
    checkEntities,
    createLocation,
    createSublocation,
    Server,
} from './helpers.mjs';

describe('Tag Associations Management API', function () {
    const tagAssociationOne = {};
    const tagAssociationTwo = {};
    const tagAssociationThree = {};
    const trackableOne = {};
    const trackableTwo = {};
    const trackableThree = {};
    const tagOne = {};
    const tagTwo = {};
    const tagThree = {};
    const anchor = {};
    const node = {};
    const nodeAssociation = {};
    const location = {};
    const sublocation = {};
    const trackableCount = 3;

    before(async function () {
        await Server().init();
        Object.assign(location, await createLocation(`Location ${UID()}`));
        Object.assign(sublocation, await createSublocation(`Sublocation ${UID()}`, location));
        Object.assign(trackableOne, await createTrackable(`Trackable ${UID()}`));
        Object.assign(trackableTwo, await createTrackable(`Trackable ${UID()}`));
        Object.assign(trackableThree, await createTrackable(`Trackable ${UID()}`));
        Object.assign(tagOne, await createTag('04956EA5010c'));
        Object.assign(tagTwo, await createTag('04956EA5010b'));
        Object.assign(tagThree, await createTag('04956EA5010a'));
        Object.assign(node, await createNode('04956EA50109'));
        Object.assign(anchor, await createAnchor(`Anchor ${UID()}`, sublocation));
        Object.assign(nodeAssociation, await createNodeAssociation(anchor, node));
    });

    after(async function () {
        await del(`/api/v3/locations/${location.uid}`); // cascade remove of sublocations/anchors
        await del(`/api/v3/trackables/${trackableOne.uid}`);
        await del(`/api/v3/trackables/${trackableTwo.uid}`);
        await del(`/api/v3/trackables/${trackableThree.uid}`);
        await del(`/api/v3/tags/${tagOne.mac}`);
        await del(`/api/v3/tags/${tagTwo.mac}`);
        await del(`/api/v3/tags/${tagThree.mac}`);
        await del(`/api/v3/nodes/${node.mac}`);
    });

    describe('Trackables before tag association creation', function () {
        it('should return list of trackables with tag_mac equal to null', function (done) {
            get(`/api/v3/trackables`)
                .expect(200)
                .then((response) => {
                    const { trackables } = response.body;
                    assert(trackables.length === trackableCount, `Wrong trackables number`);
                    for (const trackable of trackables) {
                        assert(trackable.tag_mac === null, 'trackable should have mac link equal null');
                    }
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Get Empty Tag Associations List', function () {
        it('should return empty list of tag associations', checkEntities('tag_assignments', []));
    });

    describe('Create Tag Association', function () {
        describe('Negative cases 1', function () {
            it('should fail create tag association when body with no trackable_uid', function (done) {
                const body = {
                    mac: tagOne.mac,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"trackable_uid":"REQUIRED"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association when body with empty trackable_uid', function (done) {
                const body = {
                    mac: tagOne.mac,
                    trackable_uid: '',
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"trackable_uid":"CANNOT_BE_EMPTY"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association when body without mac, REQUIRED', function (done) {
                const body = {
                    trackable_uid: trackableOne.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"mac":"REQUIRED"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association when body is empty mac, CANNOT_BE_EMPTY', function (done) {
                const body = {
                    trackable_uid: trackableOne.uid,
                    mac: '',
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"mac":"CANNOT_BE_EMPTY"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association when body with invalid MAC, WRONG_MAC_FORMAT', function (done) {
                const body = {
                    trackable_uid: trackableOne.uid,
                    mac: 'MAC12345IT22',
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"mac":"WRONG_MAC_FORMAT"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association with non existing anchor', function (done) {
                const body = {
                    mac: tagOne.mac,
                    trackable_uid: UID(),
                };

                post('/api/v3/tag_assignments', body)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Trackable with uid ${body.trackable_uid} not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association with non existing tag', function (done) {
                const body = {
                    mac: '00000000001a',
                    trackable_uid: trackableOne.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Tag with mac ${body.mac.toUpperCase()} not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association with node', function (done) {
                const body = {
                    mac: node.mac,
                    trackable_uid: trackableOne.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, `Tag with mac ${body.mac} not found`, `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
        });
        describe('Create first', function () {
            it('should create tag association', function (done) {
                const body = {
                    mac: tagOne.mac.toLowerCase(),
                    trackable_uid: trackableOne.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(200)
                    .then((response) => {
                        Object.assign(tagAssociationOne, response.body);
                        assert.strictEqual(
                            response.body.mac.toUpperCase(),
                            body.mac.toUpperCase(),
                            `Wrong tag association "mac" value`
                        );
                        assert.strictEqual(
                            response.body.trackable_uid,
                            body.trackable_uid,
                            `Wrong tag association "trackable_uid" value`
                        );
                        assert(isISOTimeFormat(response.body.created_at), `Wrong tag association "created_at" format`);
                        assert(isUpperCaseString(response.body.mac), `Wrong tag association "mac" format`);
                        done();
                    })
                    .catch((err) => done(err));
            });
        });
        describe('Negative cases 2', function () {
            it('should fail create tag association with short node MAC address', function (done) {
                const body = {
                    mac: '0000001a',
                    trackable_uid: trackableOne.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"mac":"WRONG_MAC_FORMAT"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association with too long node MAC address', function (done) {
                const body = {
                    mac: '00000000000000001a',
                    trackable_uid: trackableOne.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"mac":"WRONG_MAC_FORMAT"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association with tag in existed association', function (done) {
                const body = {
                    mac: tagOne.mac,
                    trackable_uid: trackableTwo.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Assignment with tag ${body.mac} already exist`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association with trackable in association', function (done) {
                const body = {
                    mac: tagTwo.mac,
                    trackable_uid: trackableOne.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Assignment with trackable ${body.trackable_uid} already exist`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
        });
        describe('Create next', function () {
            it('should create second tag association', function (done) {
                const body = {
                    mac: tagTwo.mac.toLowerCase(),
                    trackable_uid: trackableTwo.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(200)
                    .then((response) => {
                        Object.assign(tagAssociationTwo, response.body);
                        assert.strictEqual(
                            response.body.mac.toUpperCase(),
                            body.mac.toUpperCase(),
                            `Wrong tag association "mac" value`
                        );
                        assert.strictEqual(
                            response.body.trackable_uid,
                            body.trackable_uid,
                            `Wrong node association "trackable_uid" value`
                        );
                        assert(isISOTimeFormat(response.body.created_at), `Wrong tag association "created_at" format`);
                        assert(isUpperCaseString(response.body.mac), `Wrong tag association "mac" format`);
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should create third node association', function (done) {
                const body = {
                    mac: tagThree.mac,
                    trackable_uid: trackableThree.uid,
                };

                post('/api/v3/tag_assignments', body)
                    .expect(200)
                    .then((response) => {
                        Object.assign(tagAssociationThree, response.body);
                        assert.strictEqual(response.body.mac, body.mac, `Wrong tag association "mac" value`);
                        assert.strictEqual(
                            response.body.trackable_uid,
                            body.trackable_uid,
                            `Wrong node trackable "trackable_uid" value`
                        );
                        assert(isISOTimeFormat(response.body.created_at), `Wrong anchor "created_at" format`);
                        assert(isUpperCaseString(response.body.mac), `Response MAC address should be in uppercase`);
                        done();
                    })
                    .catch((err) => done(err));
            });
        });
    });

    describe('Trackable after tag association creation', function () {
        it('should return list of trackables with tag_mac equal to null', function (done) {
            get(`/api/v3/trackables`)
                .expect(200)
                .then((response) => {
                    const { trackables } = response.body;
                    const tagAssociations = [tagAssociationOne, tagAssociationTwo, tagAssociationThree];
                    assert(trackables.length === trackableCount, `Wrong trackables number`);
                    for (const trackable of trackables) {
                        const tagAssociation = tagAssociations.find((item) => item.trackable_uid === trackable.uid);
                        assert(trackable.tag_mac === tagAssociation.mac, 'trackable should have mac link');
                    }
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Close Tag Association Management', function () {
        describe('Negative cases 2', function () {
            it('should fails close first tag association with empty body', function (done) {
                const body = {};
                put(`/api/v3/tag_assignments`, body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `{"mac":"REQUIRED","trackable_uid":"REQUIRED"}`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fails close first tag association with empty mac', function (done) {
                const body = {
                    mac: '',
                };

                put('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"mac":"CANNOT_BE_EMPTY"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fails close first tag association with empty trackable_uid', function (done) {
                const body = {
                    trackable_uid: '',
                };

                put('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"trackable_uid":"CANNOT_BE_EMPTY"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fails close first tag association with empty mac and empty trackable_uid', function (done) {
                const body = {
                    mac: '',
                    trackable_uid: '',
                };

                put('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            '{"mac":"CANNOT_BE_EMPTY","trackable_uid":"CANNOT_BE_EMPTY"}',
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });

            it('should fail create tag association when body is empty mac, CANNOT_BE_EMPTY', function (done) {
                const body = {
                    trackable_uid: trackableOne.uid,
                    mac: '',
                };

                put('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"mac":"CANNOT_BE_EMPTY"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fail create tag association when body is wrong mac, WRONG_MAC_FORMAT', function (done) {
                const body = {
                    trackable_uid: trackableOne.uid,
                    mac: 'MAC12345IT22',
                };

                put('/api/v3/tag_assignments', body)
                    .expect(422)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(message, '{"mac":"WRONG_MAC_FORMAT"}', `Wrong error message`);
                        assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should return NotFound error non existing mac association', function (done) {
                const body = { mac: '00000000000a' };
                put(`/api/v3/tag_assignments`, body)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Assignment with tag ${body.mac.toUpperCase()} not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should fails close first tag association when correct mac wrong trackable_uid ', function (done) {
                const body = {
                    mac: tagOne.mac.toLowerCase(),
                    trackable_uid: 'sdasd',
                };

                put('/api/v3/tag_assignments', body)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Assignment with tag ${body.mac.toUpperCase()} and trackable ${
                                body.trackable_uid
                            } not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });

            it('should fails close first tag association when correct trackable_uid wrong mac ', function (done) {
                const body = {
                    mac: '000ee000002a',
                    trackable_uid: trackableOne.uid,
                };

                put('/api/v3/tag_assignments', body)
                    .expect(404)
                    .then((response) => {
                        const { message, code } = response.body;
                        assert.deepStrictEqual(
                            message,
                            `Assignment with tag ${body.mac.toUpperCase()} and trackable ${
                                body.trackable_uid
                            } not found`,
                            `Wrong error message`
                        );
                        assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                        done();
                    })
                    .catch((err) => done(err));
            });
        });
        describe('Positive cases', function () {
            it('should close tag association', function (done) {
                const body = { mac: tagAssociationOne.mac.toLowerCase() };
                put(`/api/v3/tag_assignments`, body)
                    .expect(200)
                    .then((response) => {
                        assert.strictEqual(
                            response.body.mac.toUpperCase(),
                            body.mac.toUpperCase(),
                            `Wrong tag association "mac" value`
                        );
                        assert.strictEqual(
                            response.body.trackable_uid,
                            trackableOne.uid,
                            `Wrong tag association "trackable_uid" value`
                        );
                        assert(isISOTimeFormat(response.body.created_at), `Wrong tag association "created_at" format`);
                        assert(isUpperCaseString(response.body.mac), `Wrong tag association "mac" format`);
                        done();
                    })
                    .catch((err) => done(err));
            });
            it('should re-create tag association', function (done) {
                const body = { mac: tagOne.mac.toLowerCase(), trackable_uid: trackableOne.uid };
                post(`/api/v3/tag_assignments`, body)
                    .expect(200)
                    .then((response) => {
                        Object.assign(tagAssociationOne, response.body);
                        assert.strictEqual(
                            response.body.mac.toUpperCase(),
                            body.mac.toUpperCase(),
                            `Wrong tag association "mac" value`
                        );
                        assert.strictEqual(
                            response.body.trackable_uid,
                            body.trackable_uid,
                            `Wrong tag association "trackable_uid" value`
                        );
                        assert(isISOTimeFormat(response.body.created_at), `Wrong tag association "created_at" format`);
                        assert(isUpperCaseString(response.body.mac), `Wrong tag association "mac" format`);
                        done();
                    })
                    .catch((err) => done(err));
            });
        });
    });

    describe('Clear Tag Association', function () {
        it('should return list of tag associations', function (done) {
            get('/api/v3/tag_assignments')
                .expect(200)
                .then((response) => {
                    const resAssociationOne = response.body.tag_assignments.find(
                        (assoc) => assoc.trackable_uid === tagAssociationOne.trackable_uid
                    );
                    const resAssociationTwo = response.body.tag_assignments.find(
                        (assoc) => assoc.trackable_uid === tagAssociationTwo.trackable_uid
                    );
                    const resAssociationThree = response.body.tag_assignments.find(
                        (assoc) => assoc.trackable_uid === tagAssociationThree.trackable_uid
                    );
                    assert.strictEqual(
                        response.body.tag_assignments.length,
                        3,
                        'Associations list should contain 3 items'
                    );
                    assert.deepStrictEqual(resAssociationOne, tagAssociationOne, 'Associations should be equal');
                    assert.deepStrictEqual(resAssociationTwo, tagAssociationTwo, 'Associations should be equal');
                    assert.deepStrictEqual(resAssociationThree, tagAssociationThree, 'Associations should be equal');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should close tag association one by anchor_uid', function (done) {
            const body = { trackable_uid: tagAssociationOne.trackable_uid };
            put(`/api/v3/tag_assignments`, body)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(
                        response.body.trackable_uid,
                        body.trackable_uid,
                        `Wrong tag association "mac" value`
                    );
                    assert.strictEqual(
                        response.body.trackable_uid,
                        tagAssociationOne.trackable_uid,
                        `Wrong tag association "trackable_uid" value`
                    );
                    assert(isISOTimeFormat(response.body.created_at), `Wrong tag association "created_at" format`);
                    assert(isUpperCaseString(response.body.mac), `Wrong tag association "mac" format`);
                    done();
                })
                .catch((err) => done(err));
        });

        it('should close tag association two by mac', function (done) {
            const body = { mac: tagAssociationTwo.mac.toLowerCase() };
            put(`/api/v3/tag_assignments`, body)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(
                        response.body.mac.toUpperCase(),
                        body.mac.toUpperCase(),
                        `Wrong tag association "mac" value`
                    );
                    assert.strictEqual(
                        response.body.trackable_uid,
                        tagAssociationTwo.trackable_uid,
                        `Wrong tag association "trackable_uid" value`
                    );
                    assert(isISOTimeFormat(response.body.created_at), `Wrong tag association "created_at" format`);
                    assert(isUpperCaseString(response.body.mac), `Wrong tag association "mac" format`);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should close tag association three by mac', function (done) {
            const body = {
                mac: tagAssociationThree.mac.toLowerCase(),
                trackable_uid: tagAssociationThree.trackable_uid,
            };
            put(`/api/v3/tag_assignments`, body)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(
                        response.body.mac.toUpperCase(),
                        body.mac.toUpperCase(),
                        `Wrong tag association "mac" value`
                    );
                    assert.strictEqual(
                        response.body.trackable_uid,
                        tagAssociationThree.trackable_uid,
                        `Wrong tag association "trackable_uid" value`
                    );
                    assert(isISOTimeFormat(response.body.created_at), `Wrong tag association "created_at" format`);
                    assert(isUpperCaseString(response.body.mac), `Wrong tag association "mac" format`);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should return empty list of tag associations after cleanup', checkEntities('tag_assignments', []));
    });

    describe('Trackables after tag association removing', function () {
        it('should return list of trackables with tag_mac equal to null', function (done) {
            get(`/api/v3/trackables`)
                .expect(200)
                .then((response) => {
                    const { trackables } = response.body;
                    assert(trackables.length === trackableCount, `Wrong trackables number`);
                    for (const trackable of trackables) {
                        assert(trackable.tag_mac === null, 'trackable should have mac link equal null');
                    }
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('ZMQ Messaging for Tag Assignments CRUD', function () {
        const zmqObserver = new ZmqObserver();
        const tagAssociationZmq = {};

        before(function () {
            zmqObserver.init();
        });
        after(function () {
            zmqObserver.destroy();
        });

        it('should send zmq created tag assignment message', function (done) {
            const zmqSubscription = { action: ZMQ_ACTIONS.CREATED, object: ZMQ_OBJECTS.TAG_ASSIGNMENT };
            const body = {
                mac: tagOne.mac.toLowerCase(),
                trackable_uid: trackableOne.uid,
            };

            const subscription = zmqObserver.subscribe(zmqSubscription, function (zmqMsg, channel) {
                assert(isISOTimeFormat(zmqMsg.ts), `Wrong ZMQ message "ts" format`);
                assert(channel === ZMQ_CHANNELS.EVENT, 'Wrong ZMQ message channel');
                assert(zmqMsg.data.mac === body.mac.toUpperCase(), 'ZMQ message has wrong assignment mac address');
                assert(zmqMsg.data.trackable_uid === body.trackable_uid, 'ZMQ message has wrong trackable_uid prop');
                assert(isISOTimeFormat(zmqMsg.data.created_at), 'Wrong ZMQ message created_at value');
                done();
            });
            subscription
                .then(() => post(`/api/v3/tag_assignments`, body))
                .then((response) => {
                    Object.assign(tagAssociationZmq, response.body);
                });
        });
        it('should send zmq deleted tag assignment message', function (done) {
            const zmqSubscription = { action: ZMQ_ACTIONS.DELETED, object: ZMQ_OBJECTS.TAG_ASSIGNMENT };

            const subscription = zmqObserver.subscribe(zmqSubscription, function (zmqMsg, channel) {
                assert(isISOTimeFormat(zmqMsg.ts), `Wrong ZMQ message "ts" format`);
                assert(channel === ZMQ_CHANNELS.EVENT, 'Wrong ZMQ message channel');
                assert(
                    zmqMsg.data.mac === tagAssociationZmq.mac.toUpperCase(),
                    'ZMQ message has wrong assignment mac address'
                );
                assert(
                    zmqMsg.data.trackable_uid === tagAssociationZmq.trackable_uid,
                    'ZMQ message has wrong trackable_uid prop'
                );
                assert(isISOTimeFormat(zmqMsg.data.created_at), 'Wrong ZMQ message created_at value');
                done();
            });
            subscription.then(() => put(`/api/v3/tag_assignments`, tagAssociationZmq));
        });
    });
});
