import zmq from 'zeromq/v5-compat.js';
import * as ConfigContainer from '../../lib/config.cjs';
import { ZMQ_CHANNELS } from '../../lib/system/constants.mjs';

export class ZmqObserver {
    /**
     * @typedef {Class} ZmqObserver
     */

    /**
     * @property {Object|null}
     */
    #socket = null;

    /**
     * @property {Array<Object>}
     */
    #subscriptions = [];

    constructor() {
        this.#socket = zmq.socket('sub');
        this.#socket.connect(ConfigContainer.config.zmqSubUrl);
        this.#subscriptions = [];
    }

    /**
     * @method
     * @return {VoidFunction}
     */
    init() {
        this.#socket.subscribe('');
        this.#socket.on('message', this.handleMessage);
    }

    /**
     * @method
     * @return {VoidFunction}
     */
    destroy() {
        this.#socket.close();
    }

    /**
     * @static
     * @param {Object} messageOne - zmqMessage
     * @param {Object} messageTwo - zmqMessage
     * @return {Boolean}
     */
    static messagesAreEqual(messageOne, messageTwo) {
        return messageOne.action === messageTwo.action && messageOne.object === messageTwo.object;
    }

    /**
     * @static
     * @param {Buffer} bufferMsg - Buffer from ZMQ
     * @return {Array<any>|null}
     */
    static parseZmqMsg(bufferMsg) {
        const message = bufferMsg.toString();
        if (message.startsWith(`${ZMQ_CHANNELS.EVENT} `)) {
            const result = JSON.parse(message.replace(`${ZMQ_CHANNELS.EVENT} `, ''));
            return [result, ZMQ_CHANNELS.EVENT];
        } else if (message.startsWith(`${ZMQ_CHANNELS.POS} `)) {
            const result = JSON.parse(message.replace(`${ZMQ_CHANNELS.POS} `, ''));
            return [result, ZMQ_CHANNELS.POS];
        } else if (message.startsWith(`${ZMQ_CHANNELS.DATA} `)) {
            const result = JSON.parse(message.replace(`${ZMQ_CHANNELS.DATA} `, ''));
            return [result, ZMQ_CHANNELS.DATA];
        }
        return null;
    }

    /**
     * @method
     * @param {Buffer} bufferMsg - Buffer from ZMQ
     * @return {Boolean}
     */
    handleMessage = (bufferMsg) => {
        const [zmqMsg, channel] = ZmqObserver.parseZmqMsg(bufferMsg);
        const subscription = this.#findSubscription(zmqMsg);
        if (subscription) {
            this.#unsubscribe(subscription);
            subscription.callback(zmqMsg, channel);
        }
    };

    /**
     * @method
     * @param {Object} zmqObject - parsed ZMQ object
     * @return {Object}
     */
    #findSubscription(zmqObject) {
        return this.#subscriptions.find((item) => ZmqObserver.messagesAreEqual(item.subscription, zmqObject));
    }

    /**
     * @method
     * @param {Object} subscription - subscription
     * @return {VoidFunction}
     */
    #unsubscribe(subscription) {
        this.#subscriptions = this.#subscriptions.filter((item) => item !== subscription);
    }

    /**
     * @method
     * @param {Object} subscription - subscription
     * @param {Function} callback - callback
     * @return {VoidFunction}
     */
    async subscribe(subscription = {}, callback = () => {}) {
        this.#subscriptions.push({ subscription, callback });
    }
}
