import makeRequestHandler from '../../../utils/makeRequestHandler.mjs';
import { DeviceGet, DeviceSave } from '../../../../usecases/index.mjs';

export default {
    /**
     * @swagger
     * /api/v3/hardware_devices:
     *   get:
     *     description: Retrieve auto-created devices list
     *     tags:
     *       - Devices API
     *     responses:
     *       '200':
     *         description: tags
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/Devices'
     *              example:
     *                  hardware_devices:
     *                      -
     *                          mac: E4956EA5010A
     *                          last_heard: 2021-04-21T10:00:00.000Z
     *                          node_id: 1
     *                          cbid: null
     *                          created_at: 2021-04-21T10:00:00.000Z
     *                      -
     *                          mac: E4956EA501BB
     *                          last_heard: 2021-04-21T10:00:00.000Z
     *                          node_id: 2
     *                          cbid: null
     *                          created_at: 2021-04-21T10:00:00.000Z
     */
    getAll: makeRequestHandler(
        DeviceGet,
        (req) => ({}),
        (devices, res) => {
            res.json({ hardware_devices: devices });
        }
    ),

    /**
     * @swagger
     * /api/v3/hardware_devices/{mac}:
     *   get:
     *     description: Retrieve auto-created device by MAC address
     *     tags:
     *       - Devices API
     *     parameters:
     *      - in: path
     *        name: mac
     *        schema:
     *          type: string
     *          required: true
     *          example: e4956ea5010a
     *          description: device MAC address
     *     responses:
     *       '200':
     *         description: devices
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/Devices'
     *              example:
     *                  hardware_devices:
     *                      -
     *                          mac: E4956EA5010A
     *                          last_heard: 2021-04-21T10:00:00.000Z
     *                          node_id: 1
     *                          cbid: null
     *                          created_at: 2021-04-21T10:00:00.000Z
     *       '404':
     *         description: An device with the specified MAC was not found.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NotFoundError'
     *              example:
     *                  message: Device with MAC ... not found
     *                  code: NOT_FOUND_ERROR
     *       '422':
     *         description: Wrong format of the device MAC address.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/ValidationError'
     *              example:
     *                  message: '{"mac":"WRONG_MAC_FORMAT"}'
     *                  code: VALIDATION_ERROR
     */
    get: makeRequestHandler(
        DeviceGet,
        (req) => ({
            mac: req.params && req.params.mac,
        }),
        (devices, res) => {
            res.json({ hardware_devices: devices });
        }
    ),

    /**
     * /api/v3/hardware_devices: # not in documentation
     *   post:
     *     description: Create device
     *     tags:
     *       - Devices API
     *     requestBody:
     *        content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/Devices'
     *              example:
     *                  mac: e4956ea5010a
     *     responses:
     *       '200':
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/Devices'
     *              example:
     *                  mac: E4956EA5010A
     *                  last_heard: 2021-04-21T10:00:00.000Z
     *                  node_id: 1
     *                  cbid: null
     *                  created_at: 2021-04-21T10:00:00.000Z
     *       '422':
     *          description: Invalid request parameters.
     *          content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/ValidationError'
     *                  code: VALIDATION_ERROR
     *              examples:
     *                  No MAC specified:
     *                      value:
     *                          message: '{"mac":"REQUIRED"}'
     *                          code: VALIDATION_ERROR
     *                  Empty mac:
     *                      value:
     *                          message: '{"mac":"CANNOT_BE_EMPTY"}'
     *                          code: VALIDATION_ERROR
     *                  Wrong MAC format:
     *                      value:
     *                          message: '{"mac":"WRONG_MAC_FORMAT"}'
     *                          code: VALIDATION_ERROR
     *                  MAC not available:
     *                      value:
     *                          message: Node or tag with MAC ... already exist
     *                          code: VALIDATION_ERROR
     */
    post: makeRequestHandler(
        DeviceSave,
        (req) => ({
            mac: req.body.mac,
        }),
        (result, res) => {
            res.json(result);
        }
    ),
};
