import ModelBase from '../ModelBase.mjs';
import { NotFoundError } from '../errors/index.mjs';
import { UID } from '../../utils/index.mjs';
import { formatAnchorResponse, formatAnchorResponseList } from './formatters.mjs';

/**
 * @swagger
 * components:
 *  schemas:
 *      AnchorCreateBody:
 *          type: object
 *          properties:
 *              name:
 *                  type: string
 *                  required: true
 *              bridge:
 *                  type: boolean
 *                  required: true
 *              position:
 *                  type: object
 *                  required: true
 *                  properties:
 *                      x:
 *                          type: number
 *                          required: true
 *                      y:
 *                          type: number
 *                          required: true
 *                      z:
 *                          type: number
 *                          required: true
 *                      sublocation_uid:
 *                          type: string
 *                          required: true
 *      AnchorUpdateBody:
 *          type: object
 *          properties:
 *              name:
 *                  type: string
 *                  required: false
 *              bridge:
 *                  type: boolean
 *                  required: false
 *              position:
 *                  type: object
 *                  required: false
 *                  properties:
 *                      x:
 *                          type: number
 *                          required: true
 *                      y:
 *                          type: number
 *                          required: true
 *                      z:
 *                          type: number
 *                          required: true
 *                      sublocation_uid:
 *                          type: string
 *                          required: true
 *      Anchor:
 *          type: object
 *          properties:
 *              uid:
 *                  type: string
 *                  description: Anchor unique identifier.
 *              name:
 *                  type: string
 *                  description: Anchor name.
 *              bridge:
 *                  type: boolean
 *                  description: The flag that indicates is the anchor has a BRIDGE role.
 *              created_at:
 *                  type: string (ISO 8601)
 *                  description: The time stamp when the anchor was created, in ISO 8601 format.
 *              updated_at:
 *                  type: string (ISO 8601)
 *                  description: The time stamp when the anchor was updated, in ISO 8601 format.
 *              node_mac:
 *                  type: string
 *                  description: The MAC address of the associated hardware device.
 *              position:
 *                  type: object
 *                  description: JSON object indicates the anchor position.
 *                  properties:
 *                      x:
 *                          type: number
 *                          description: The X coordinate of the anchor position.
 *                      y:
 *                          type: number
 *                          description: The Y coordinate of the anchor position.
 *                      z:
 *                          type: number
 *                          description: The Z coordinate of the anchor position.
 *                      sublocation_uid:
 *                          type: string
 *                          description: Unique identifier of the position sublocation.
 *      Anchors:
 *          type: object
 *          properties:
 *              anchors:
 *                  type: array
 *                  items:
 *                      $ref: '#/components/schemas/Anchor'
 */

export class Anchor extends ModelBase {
    /**
     * @property {String|null}
     */
    static object = 'anchor';

    /**
     * @method
     * @param {Object} params - anchor properties to save
     * @param {String} params.name - anchors name
     * @param {Boolean} params.bridge - the flag that indicates is the anchor has a BRIDGE role
     * @param {Object} params.position - JSON object indicates the anchor position
     * @param {Number} params.position.x - The X coordinate of the anchor position
     * @param {Number} params.position.y - The Y coordinate of the anchor position
     * @param {Number} params.position.z - The Z coordinate of the anchor position
     * @param {String} params.position.sublocation_uid - Unique identifier of the position sublocation
     * @return {Promise<Object>}
     */
    async save(params) {
        const sublocation = await this.repository.sublocation.findByUid(params.position.sublocation_uid);

        if (!sublocation) {
            throw new NotFoundError(`Sublocation with uid ${params.position.sublocation_uid} not found`);
        }
        const now = new Date().toISOString();
        const anchor = {
            uid: UID(),
            created_at: now,
            updated_at: now,
            ...params,
        };
        const result = await this.repository.anchor.save(anchor);
        return formatAnchorResponse(result);
    }

    /**
     @typedef AnchorWithChildren
     @type {Object}
     @property {Object} anchor - Anchor.
     @property {Object} nodeAssociation - Node association.
     */
    /**
     * @method
     * @param {Object} params - options for read anchor in repository
     * @param {String} params.uid - anchor unique identifier
     * @return {Promise<AnchorWithChildren>}
     */
    async loadWithAssociation(params) {
        const { uid } = params;
        const anchor = await this.repository.anchor.findByUid(uid);
        if (!anchor) {
            throw new NotFoundError(`Anchor with uid ${uid} not found`);
        }
        const nodeAssociation = await this.repository.association.findByObjUid(uid);
        return { anchor, nodeAssociation };
    }

    /**
     * @method
     * @param {Object} params - options for read anchor in repository
     * @param {String|undefined} params.uid - anchor unique identifier
     * @return {Promise<Object>}
     */
    async load(params) {
        const { uid } = params;
        if (!uid) {
            const dbAnchors = (await this.repository.anchor.findAll()) || [];
            const nodeAssociations = (await this.repository.association.findAll()) || [];
            const anchorsWithNodeMac = dbAnchors.reduce((memo, anchor) => {
                const nodeAssociation = nodeAssociations.find((assoc) => assoc.anchor_uid === anchor.uid);
                if (nodeAssociation) {
                    memo.push({ ...anchor, node_mac: nodeAssociation.mac });
                } else {
                    memo.push(anchor);
                }
                return memo;
            }, []);
            const anchors = formatAnchorResponseList(anchorsWithNodeMac);

            return { anchors };
        }

        const anchor = await this.repository.anchor.findByUid(uid);
        if (!anchor) {
            throw new NotFoundError(`Anchor with uid ${uid} not found`);
        }

        const nodeAssociation = await this.repository.association.findByObjUid(anchor.uid);
        if (nodeAssociation) {
            anchor.node_mac = nodeAssociation.mac;
        }

        return { anchors: [formatAnchorResponse(anchor)] };
    }

    /**
     * @method
     * @param {Object} params - options for read anchor in repository
     * @param {String} params.uid - anchor unique identifier
     * @return {Promise<Object>}
     */
    async remove(params) {
        const { uid } = params;
        const anchor = await this.repository.anchor.findByUid(uid);
        if (!anchor) {
            throw new NotFoundError(`Anchor with uid ${uid} not found`);
        }
        await this.repository.anchor.deleteByUid(uid);

        return formatAnchorResponse(anchor);
    }

    /**
     * @method
     * @param {Object} params - anchor properties to update
     * @param {String} params.uid - anchor unique identifier
     * @param {String|undefined} params.name - anchor name
     * @param {Boolean|undefined} params.bridge - the flag that indicates is the anchor has a BRIDGE role
     * @param {Object|undefined} params.position - JSON object indicates the anchor position
     * @param {Number|undefined} params.position.x - The X coordinate of the anchor position
     * @param {Number|undefined} params.position.y - The X coordinate of the anchor position
     * @param {Number|undefined} params.position.z - The X coordinate of the anchor position
     * @param {String|undefined} params.position.sublocation_uid - Unique identifier of the position sublocation
     * @return {Promise<Object>}
     */
    async update(params) {
        const { uid, position } = params;
        const anchor = await this.repository.anchor.findByUid(uid);
        if (!anchor) {
            throw new NotFoundError(`Anchor with uid ${uid} not found`);
        }
        if (position && position.sublocation_uid) {
            const sublocation = await this.repository.sublocation.findByUid(position.sublocation_uid);

            if (!sublocation) {
                throw new NotFoundError(`Sublocation with uid ${position.sublocation_uid} not found`);
            }
        }

        await this.repository.anchor.save({ ...params, updated_at: new Date().toISOString() });

        const updatedAnchor = await this.repository.anchor.findByUid(uid);

        const nodeAssociation = await this.repository.association.findByObjUid(updatedAnchor.uid);
        if (nodeAssociation) {
            updatedAnchor.node_mac = nodeAssociation.mac;
        }

        return formatAnchorResponse(updatedAnchor);
    }
}
