import ModelBase from '../ModelBase.mjs';
import { NODE_TYPE } from '../../system/constants.mjs';
import { NotFoundError, ValidationError } from '../errors/index.mjs';

/**
 * @swagger
 * components:
 *  schemas:
 *      NodeAssociationCreateBody:
 *          type: object
 *          properties:
 *              mac:
 *                  type: string
 *                  required: true
 *              anchor_uid:
 *                  type: string
 *                  required: true
 *      NodeAssociationCloseBody:
 *          type: object
 *          properties:
 *              mac:
 *                  type: string
 *                  required: false
 *              anchor_uid:
 *                  type: string
 *                  required: false
 *      NodeAssignment:
 *          type: object
 *          properties:
 *              anchor_uid:
 *                  type: string
 *                  description: The unique identifier of the assigned anchor.
 *              mac:
 *                  type: string
 *                  description: The MAC address of the assigned node. In upper case.
 *              created_at:
 *                  type: string (ISO 8601)
 *                  description: The time stamp when the node association was created, in ISO 8601 format.
 *      NodeAssignments:
 *          type: object
 *          properties:
 *              node_assignments:
 *                  type: array
 *                  items:
 *                      $ref: '#/components/schemas/NodeAssignment'
 */

export class NodeAssociation extends ModelBase {
    #entityType = NODE_TYPE.WIN;

    /**
     * @property {String|null}
     */
    static object = 'node_assignment';

    /**
     * @method
     * @param {Object} params - association properties to save
     * @param {String} params.mac - node mac address
     * @param {String} params.anchor_uid - anchor unique identifier
     * @return {Promise<Object>}
     */
    async save(params) {
        const now = new Date().toISOString();
        const anchor = await this.repository.anchor.findByUid(params.anchor_uid);

        if (!anchor) {
            throw new NotFoundError(`Anchor with uid ${params.anchor_uid} not found`);
        }

        const node = await this.repository.node.findByMac(params.mac);

        if (!node || node.type !== this.#entityType) {
            throw new NotFoundError(`Node with mac ${params.mac} not found`);
        }

        const associationWithNode = await this.repository.association.findByMac(params.mac);

        if (associationWithNode) {
            throw new ValidationError(`Assignment with node ${params.mac} already exist`);
        }

        const associationWithAnchor = await this.repository.association.findByObjUid(params.anchor_uid);

        if (associationWithAnchor) {
            throw new ValidationError(`Assignment with anchor ${params.anchor_uid} already exist`);
        }

        const nodeAssociation = {
            created_at: now,
            ...params,
        };

        return this.repository.association.save(nodeAssociation);
    }

    /**
     * @method
     * @param {Object} params - association properties
     * @param {String|undefined} params.uid - anchor unique identifier
     * @return {Promise<Object>}
     */
    async load(params = {}) {
        if (params.uid) {
            const association = await this.repository.association.findByObjUid(params.uid);
            if (!association.anchor_uid) {
                throw new NotFoundError(`Assignment with anchor ${params.uid} not found`);
            }
        }
        return ((await this.repository.association.findAll()) || []).filter((association) => association.anchor_uid);
    }

    /**
     * @method
     * @param {Object} params - association properties
     * @param {String} params.mac - node mac address
     * @param {String} params.anchor_uid - unique id of anchor
     * @return {Promise<Object>}
     */
    async close(params) {
        const { mac, anchor_uid } = params;

        let association;
        if (mac && anchor_uid) {
            association = await this.repository.association.findByMac(mac);
            if (!association || (association && association.anchor_uid !== anchor_uid)) {
                throw new NotFoundError(`Assignment with node ${mac} and anchor ${anchor_uid} not found`);
            }
        } else if (mac) {
            association = await this.repository.association.findByMac(mac);
            if (!association) {
                throw new NotFoundError(`Assignment with node ${mac} not found`);
            }
        } else if (anchor_uid) {
            association = await this.repository.association.findByObjUid(anchor_uid);
            if (!association) {
                throw new NotFoundError(`Assignment with anchor ${anchor_uid} not found`);
            }
        }

        await this.repository.association.del(association);

        return association;
    }
}
