import { createLogger, LoggerTypes } from '@rtls-platform/logger';
import * as PlatformRepository from '@rtls-platform/repository';

import * as ConfigContainer from './config/config.cjs';
import { setLogger } from './lib/logger.mjs';
import * as SolutionRepository from './lib/solutionRepository/index.mjs';
import { createDbConnection } from './lib/db/redisClient.mjs';
import { SolutionDBPoolFactory } from './lib/db/solutionDBClient.js';
import { syncEntities } from './lib/index.mjs';
import ModelBase from './lib/models/ModelBase.mjs';

/**
 *
 * @return {Promise<{repository: *}>}
 */
async function execute() {
    const platformRepositoryInstance = PlatformRepository.createRepository({
        repositoryInstance: createDbConnection({
            port: ConfigContainer.config.db.port,
            host: ConfigContainer.config.db.host
        }),
        prefix: ConfigContainer.config.db.prefix,
        shared: true,
        nodeIdLimit: ConfigContainer.config.nodeIdLimit,
        networkIdLimit: ConfigContainer.config.networkIdLimit,
        sublocationIdLimit: ConfigContainer.config.sublocationIdLimit
    });
    const solutionRepositoryInstance = SolutionRepository.createRepository({
        repositoryInstance: SolutionDBPoolFactory.createSolutionDBPool(ConfigContainer.config.solutionDb)
    });
    ModelBase.setSolutionRepository(solutionRepositoryInstance);
    ModelBase.setPlatformRepository(platformRepositoryInstance);

    await syncEntities();

    return {
        platformRepository: platformRepositoryInstance,
    }
}

/**
 *
 * @return {Promise<{repository: *}>}
 */
export async function main() {
    const logger = createLogger({
        type: LoggerTypes.Winston,
        context: {
            transport: ConfigContainer.config.logger.transport,
            rotatePolicy: ConfigContainer.config.logger.rotatePolicy,
            logLevel: ConfigContainer.config.logLevel,
            serviceName: ConfigContainer.config.serviceName
        }
    });
    setLogger(logger);

    logger.info('Data Migration Script Start');

    async function exit() {
        logger.info('Data Migration Script Exit');
        process.exit(0);
    }

    process.on('SIGTERM', async () => {
        logger.error('SIGTERM signal caught');
        await exit();
    });

    process.on('SIGINT', async () => {
        logger.error('SIGINT signal caught');
        await exit();
    });

    process.on('unhandledRejection', (error) => {
        logger.error('unhandledRejection', error.stack);
    });

    process.on('uncaughtException', (error) => {
        logger.error('uncaughtException', error.stack);
    });

    const result = await execute();
    logger.info('Data Migration Done.');
    return result;
}

if (process.argv.includes('main')) {
    main().then(() => {
        setTimeout(()=> { process.exit(0); }, 100)
    }).catch((err) => {
        console.error(err);
        process.exit(1);
    });
}
