import winston from 'winston';
import path from 'path';
import { CoreLogger } from '../core/index.mjs';
import { objectStringify } from '../utils/index.mjs';

// NOTE: require to use winston.transports.DailyRotateFile
// eslint-disable-next-line no-unused-vars
import WinstonDailyRotate from 'winston-daily-rotate-file';


const TRANSPORTS = {
    CONSOLE: 'console',
    FILE: 'file',
    NONE: 'none',
}

const ROTATE_POLICY = {
    DAILY: 'daily',
    MONTHLY: 'monthly',
    NONE: 'none',
}

function getDatePattern(rotatePolicy) {
    if (rotatePolicy === ROTATE_POLICY.DAILY) { return 'YYYY-MM-DD' }
    if (rotatePolicy === ROTATE_POLICY.MONTHLY) { return 'YYYY-MM' }
    return 'YYYY-MM-DD';
}
/**
 * @class WinstonLogger
 * @description create winston logger that can write logs into console transport, or daily rotate transport
 */
export class WinstonLogger extends CoreLogger {
    #logger = null;

    #transports = null;

    /**
     *
     * @param params
     * @property logLevel {String}
     * @property serviceName {String}
     * @property transport {String} - {none|console|file|console,file|file,console} default console
     * @property rotatePolicy {String} - {none|daily|monthly} default none
     */
    constructor({ logLevel = 'info', serviceName = 'rtls-platform', transport = TRANSPORTS.CONSOLE, rotatePolicy = ROTATE_POLICY.NONE }) {
        super();

        const options = {
            format: winston.format.combine(
                winston.format.colorize({ all: true }),
                winston.format.timestamp({ format: 'YYYY/MM/DD HH:mm:ss' }),
                winston.format.printf((info) => `[${info.timestamp}] [${info.service}] [${info.level}] ${info.message}`)
            ),
            defaultMeta: { service: serviceName },
            transports: [new winston.transports.Console({ level: logLevel })],
        };

        this.#transports = {};
        const transportTypes = transport.split(',');
        
        const transports = [];
        
        this.#transports.console = new winston.transports.Console({ silent: !transportTypes.includes(TRANSPORTS.CONSOLE) });
        
        transports.push(this.#transports.console);
        
        if (transportTypes.includes(TRANSPORTS.FILE)) {
            if (rotatePolicy === ROTATE_POLICY.NONE) {
                // no rotation
                this.#transports.file = new winston.transports.File({
                    name: 'fileTransport',
                    filename: path.join(`/var/log/rtls/${serviceName}.log`),
                    prepend: false,
                    json: false,
                    colorize: true,
                    prettyPrint: true,
                });
                transports.push(this.#transports.file);
            } else {
                // daily / monthly rotate
                this.#transports.rotate = new winston.transports.DailyRotateFile({
                    name: 'dailyRotateTransport',
                    filename: path.join(`/var/log/rtls/${serviceName}-%DATE%.log`),
                    datePattern: getDatePattern(rotatePolicy),
                    prepend: false,
                    json: false,
                    colorize: true,
                    prettyPrint: true,
                });
                transports.push(this.#transports.rotate);
            }
        }
        this.#logger = winston.createLogger({ ...options, transports });
    }

    level({ level = 'debug' }) {
        this.#transports.console.level = level;
        this.#transports.rotate.level = level;
    }

    info(...args) {
        this.#logger.info(args.map(objectStringify).join(' '));
    }

    debug(...args) {
        this.#logger.debug(args.map(objectStringify).join(' '));
    }

    warn(...args) {
        this.#logger.warn(args.map(objectStringify).join(' '));
    }

    error(...args) {
        this.#logger.error(args.map(objectStringify).join(' '));
    }
}
