import flat from 'flat';
import { RepoError } from '../models/index.mjs';
import RepoRedisBase from '../RepoRedisBase.mjs';

const flatten = flat.flatten;
const unflatten = flat.unflatten;

/**
 * @typedef PlatformAssociation {Object}
 * @property trackable_uid {String|undefined}
 * @property anchor_uid {String|undefined}
 * @property mac {String}
 * @property created_at {String} ISO date
 */

export class AssocRepository extends RepoRedisBase {
    /**
     * @typedef {Class} AssocRepository
     */
    assocResource;
    ASSOC_LOCK_TTL = 5000;
    REDLOCK_OFF = true;

    /**
     * @param {Object} options - repository options
     * @param {Object} options.repositoryInstance - database connection
     * @param {String} options.prefix - database prefix
     */
    constructor(options) {
        const { prefix, repositoryInstance } = options;
        super(repositoryInstance, `${prefix}:assoc`);
        this.assocResource = `${prefix}:locks:assoc`;
    }

    /**
     * @method
     * @param {String} mac - node mac address
     * @return {Promise<PlatformAssociation>} returns result of search by key
     */
    async findByMac(mac) {
        const assoc = await this.repoInstance.hgetall(`${this.document}:mac:${mac}`);

        if (!assoc) return null;

        return unflatten(assoc);
    }

    /**
     * @method
     * @param {String} uid - object unique identifier
     * @return {Promise<PlatformAssociation>} returns result of search by key
     */
    async findByObjUid(uid) {
        const assoc = await this.repoInstance.hgetall(`${this.document}:obj_uid:${uid}`);

        if (!assoc) return null;

        return unflatten(assoc);
    }

    /**
     * @method
     * @return {Promise<Array<PlatformAssociation>>} returns associations list
     */
    async findAll() {
        const result = [];
        try {
            const keys = await this.scanAll('0', `${this.document}:obj_uid:*`, new Set());
            for (const key of keys) {
                const obj = await this.getByKey(key);
                if (obj) {
                    result.push(obj);
                }
            }
            return result;
        } catch (e) {
            throw new RepoError(e);
        }
    }

    /**
     * @method
     * @param {Object} association - data to save in DB
     * @return {Promise<PlatformAssociation>} returns data saved in DB
     */
    async save(association) {
        if (!this.document) {
            throw new RepoError('No document');
        }
        if (!association || typeof association !== 'object') {
            throw new RepoError(`Object saving is aborted due to invalid type: ${association}`);
        }
        const fields = [];
        for (const [key, value] of Object.entries(flatten(association))) {
            if (value) {
                // do not save empty values
                fields.push(key, value);
            }
        }
        return await this.repoInstance.executeWithLock(this.assocResource, async () => {
            const obj_uid = association.anchor_uid || association.trackable_uid;
            await this.repoInstance.hmset(`${this.document}:obj_uid:${obj_uid}`, ...fields);
            await this.repoInstance.hmset(`${this.document}:mac:${association.mac}`, ...fields);
            return association;
        });
    }

    /**
     *
     * @param {Object} association
     * @property {String} association.mac
     * @property {String|undefined} association.anchor_uid
     * @property {String|undefined} association.trackable_uid
     * @returns {Promise<void>}
     */
    async del(association) {
        if (!this.document) {
            throw new RepoError('No document');
        }
        if (!association || !association.mac || typeof association !== 'object') {
            throw new RepoError(`Object saving is aborted due to invalid type: ${association}`);
        }

        const obj_uid = association.anchor_uid || association.trackable_uid;
        return await this.repoInstance.executeWithLock(this.assocResource, async () => {
            await this.repoInstance.del(`${this.document}:obj_uid:${obj_uid}`);
            await this.repoInstance.del(`${this.document}:mac:${association.mac}`);
        });
    }
}
