import RepoRedisBase from '../RepoRedisBase.mjs';
import { RepoError } from '../models/index.mjs';

export class IdBaseRepository extends RepoRedisBase {
    /**
     * @typedef {Class} IdBaseRepository
     */

    #maxIdValue;
    /**
     * @param {Object} options - repository options
     * @param {Object} options.repositoryInstance - database connection
     * @param {String} options.prefix - database prefix
     * @param {String} entity - entity name
     */
    constructor(options, entity) {
        const { prefix, repositoryInstance, maxIdValue } = options;
        super(repositoryInstance, `${prefix}:${entity}`);
        this.#maxIdValue = maxIdValue;
        this.prefix = prefix;
    }

    /**
     *
     * @returns {Promise<number>}
     */
    async getCurrent() {
        return this.repoInstance.get(`${this.document}:value`);
    }

    /**
     *
     * @returns {Promise<number>}
     */
    async increment() {
        return this.repoInstance.incr(`${this.document}:value`);
    }

    /**
     *
     * @returns {Promise<number>}
     */
    async decrement() {
        return this.repoInstance.decr(`${this.document}:value`);
    }

    /**
     * Pop available value
     * @returns {Promise<*>}
     */
    async popAvailable() {
        return this.repoInstance.spop(`${this.document}:available`);
    }

    /**
     * Add available value
     * @param value
     * @returns {Promise<*>}
     */
    async addAvailable(value) {
        return this.repoInstance.sadd(`${this.document}:available`, value);
    }

    /**
     *
     * @return {Promise<number|number|*>}
     */
    async nextValue() {

        let idValue = await this.popAvailable();
        if (idValue) {
            return idValue;
        }
        idValue = await this.getCurrent();
        idValue = idValue < this.#maxIdValue ? await this.increment() : 0;
        if (idValue) {
            if (this.#maxIdValue >= idValue) {
                return idValue;
            }
        }

        throw new RepoError(`Id limit exceeded. MAX id value is equal to ${this.#maxIdValue}.`);
    }

}
