import flat from 'flat';
import { RepoError } from '../models/index.mjs';
import RepoRedisBase from '../RepoRedisBase.mjs';

const flatten = flat.flatten;
const unflatten = flat.unflatten;

/**
 * @typedef PlatformNode {Object}
 * @property uid {String}
 * @property mac {String}
 * @property last_heard {String} ISO date
 * @property created_at {String} ISO date
 * @property updated_at {String} ISO date
 * @property node_id {Number}
 * @property type {String}
 */

export class NodeRepository extends RepoRedisBase {
    #document = null;

    constructor(options) {
        const { prefix, repositoryInstance } = options;
        super(repositoryInstance, `${prefix}:node`);
        this.#document = `${prefix}:node`;
        this.assocResource = `${prefix}:locks:node`;
    }

    /**
     * @method
     * @return {Promise<PlatformNode[]>} returns nodes list
     */
    async findAll() {
        const result = [];
        try {
            const keys = await this.scanAll('0', `${this.#document}:uid:*`, new Set());
            for (const key of keys) {
                const obj = await this.getByKey(key);
                if (obj) {
                    result.push(obj);
                }
            }
            return result;
        } catch (e) {
            throw new RepoError(e);
        }
    }

    /**
     * @method
     * @param {String} uid - node unique identifier
     * @return {Promise<PlatformNode>} returns result of search by key
     */
    async findByUid(uid) {
        const flatten_obj = await this.repoInstance.hgetall(`${this.#document}:uid:${uid}`);
        if (!flatten_obj) {
            return null;
        }
        return unflatten(flatten_obj);
    }

    /**
     * @method
     * @param {String} mac - node mac address
     * @return {Promise<PlatformNode>} returns result of search by key
     */
    async findByMac(mac) {
        const nodeUid = await this.repoInstance.get(`${this.#document}:mac:${mac}`);
        if (!nodeUid) return null;

        const node = await this.repoInstance.hgetall(`${this.#document}:uid:${nodeUid}`);

        if (!node) {
            return null;
        }
        return unflatten(node);
    }
    
    /**
     * @method
     * @param {String} mac - node mac address
     * @return {Promise<String|null>} returns node uid
     */
    async findUidByMac(mac) {
        const nodeUid = await this.repoInstance.get(`${this.#document}:mac:${mac}`);
        if (!nodeUid) return null;
        return nodeUid;
    }

    /**
     * @method
     * @param {Object} node - data to save in DB
     * @return {Promise<PlatformNode>} returns data saved in DB
     */
    async save(node) {
        if (!this.#document) {
            throw new RepoError('No document');
        }
        if (!node || !node.uid || typeof node !== 'object') {
            throw new RepoError(`Object saving is aborted due to invalid type: ${node}`);
        }
        const fields = [];
        for (const [key, value] of Object.entries(flatten(node))) {
            if (value) {
                // do not save empty values
                fields.push(key, value);
            }
        }

        await this.repoInstance.hmset(`${this.#document}:uid:${node.uid}`, ...fields);
        await this.repoInstance.set(`${this.#document}:mac:${node.mac}`, node.uid);

        return node;
    }

    /**
     *
     * @param {Object} node
     * @property {String} node.uid
     * @property {String} node.mac
     * @returns {Promise<void>}
     */
    async del(node) {
        if (!this.#document) {
            throw new RepoError('No document');
        }
        if (!node || !node.uid || typeof node !== 'object') {
            throw new RepoError(`Object removing is aborted due to invalid type: ${node}`);
        }

        await this.repoInstance.del(`${this.#document}:uid:${node.uid}`);
        await this.repoInstance.del(`${this.#document}:mac:${node.mac}`);
    }
    
    /** @method check device if exist and create if does not
     * @param {String} mac  - hardware device MAC address
     * @param {String} ts  - timestamp in ISO 8601 format
     * @param {String} uid  - unique id
     * @param {Function} nodeIdGetter Promise returns nodeId
     * @returns {Promise<PlatformNode>}
     */
    async findByMacOrCreate(mac, ts, uid, nodeIdGetter) {
        return await this.repoInstance.executeWithLock(this.assocResource, async () => {
            const node = await this.findByMac(mac);
            if (!node) {
                const data = {
                    uid,
                    mac,
                    last_heard: ts,
                    created_at: ts,
                    node_id: await nodeIdGetter(),
                };
                await this.save(data);
                return data;
            }
            return node;
        });
    }
}
